#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#define NPOLYLINE 100
typedef struct Polyline {
  struct Polyline * next;
  int n;
  XPoint pts[NPOLYLINE];
} Polyline;
  

typedef struct motion_data_t {
  GC            gcid;
  Time          time;
  Polyline *    line;
  Polyline *    curr;
} motion_data_t;

/* functions */
GC motion_makegc ();
void motion_resize ();
void motion_redraw ();
void motion_ptr ();
void motion_keydown ();
Widget create_motion_widget ();


/* default resources */
String defresources[] = { 
  "*motion.width:  300",
  "*motion.height: 200",
  NULL,
};

/* application-specific command line options */
XrmOptionDescRec options [] = {
  { "-message","*message", XrmoptionSepArg, NULL }};

main (argc, argv)
  int    argc;
  char  *argv[];
{
  Widget        toplevel;
  Widget        motion_widget;
  XtAppContext  app;

  /*
   * Initialize the intrinsics
   */
  toplevel = XtAppInitialize (&app, "Motion", 
			      options, XtNumber(options),
			      &argc, argv, 
			      defresources, NULL, 0);

  /*
   * create the workspace widget
   */

  motion_widget = create_motion_widget ( "motion", toplevel, NULL, 0);

  /* realize the widget and enter the main loop */
  XtRealizeWidget (toplevel);
  XtAppMainLoop(app);
}

Widget
create_motion_widget (name, parent, args, num_args)
    String     name;
    Widget     parent;
    ArgList    args;
    Cardinal   num_args;
{
  Widget          motion_widget;
  motion_data_t * motion_data;

  motion_widget = XtCreateManagedWidget 
                   (name,
		    widgetClass,
		    parent, args, num_args );
  
  motion_data = (motion_data_t * ) calloc (1, sizeof (motion_data_t));

  /* expose event handler */
  XtAddEventHandler ( motion_widget, ExposureMask, 
       False, motion_redraw, motion_data);
  /* mouse buttons press/drag/release */
  XtAddEventHandler (motion_widget,
		     ButtonPressMask | 
		       ButtonMotionMask |
		       ButtonReleaseMask | 
		       PointerMotionHintMask,
		     False, motion_ptr, motion_data);
  /* key press handler */
  XtAddEventHandler ( motion_widget, KeyPressMask,
       False, motion_keydown, motion_data);
  return motion_widget;

}

/* ARGSUSED */
void 
motion_redraw (widget, data, ev )
  Widget widget;
  motion_data_t * data;
  XEvent * ev;
{
  Display * dpy = XtDisplay (widget);
  Window w      = XtWindow  (widget);
  GC gc         = motion_makegc (widget, data);
  Polyline * line;
  /* draw saved lines */
  line = data->line;
  while (line) {
    XDrawLines (dpy, w, gc, line->pts, line->n,
		CoordModeOrigin);
    line = line->next;
  }

  /* draw current line */
  line = data->curr;
  if (line) {
    XDrawLines (dpy, w, gc, line->pts, line->n,
		CoordModeOrigin);
  }
}

static XtResource resources[] = {
  {XtNforeground, XtCForeground, XtRPixel, sizeof (int),
     XtOffset(XGCValues *, foreground), 
     XtRString, "Black"}, 
  {XtNbackground, XtCBackground, XtRPixel, sizeof (int),
     XtOffset(XGCValues *, background),
     XtRString, "White"}
};


GC 
motion_makegc ( widget, data )
  Widget widget;
  motion_data_t * data;
{
  XGCValues       gcv;

  if (!data->gcid) {

    /* initialize */
    Display * dpy = XtDisplay (widget);
    Window win    = XtWindow  (widget);

    XtGetApplicationResources (widget, &gcv, resources,
                 XtNumber (resources), NULL, 0);

    /* make and initialize the drawing GC */
    data->gcid = XCreateGC (dpy, win, 0, 0);
    XSetForeground (dpy, data->gcid,
		    gcv.foreground);
    XSetBackground (dpy, data->gcid, 
		    gcv.background);

  }
  return data->gcid;
}

void motion_ptr ( widget, data, ev )
  Widget widget;
  motion_data_t * data;
  XEvent * ev;
{
  XTimeCoord * hist = NULL;
  XTimeCoord * hptr;
  XTimeCoord * hnxt;
  XTimeCoord * hend;
  XTimeCoord   nohist;
  int          nhist;
  int          n;
  Display *    dpy = XtDisplay (widget);
  Window       w   = XtWindow  (widget);
  GC           gc  = motion_makegc (widget, data);
  static GC cross = NULL;

  if (!cross) {
    cross = XCreateGC (dpy, ev->xbutton.root, 0,0);
    XSetSubwindowMode (dpy, cross, IncludeInferiors);
    XSetForeground (dpy, cross, 1L);
  }


  switch (ev->type) {
  case ButtonPress:
    XDrawString (dpy, ev->xbutton.root, cross, 
		 ev->xbutton.x_root, ev->xbutton.y_root, "P",1);
    if (!data->curr) {
      data->curr = (Polyline *) malloc (sizeof(Polyline));
      data->curr->pts[0].x = ev->xbutton.x;
      data->curr->pts[0].y = ev->xbutton.y;
      data->time = ev->xbutton.time;
      data->curr->n = 1;
    }
    break;

  case MotionNotify:
    XDrawString (dpy, ev->xmotion.root, cross, 
		 ev->xmotion.x_root, ev->xmotion.y_root, "M",1);
    if (!data->curr) break;
    if ((ev->xmotion.time - data->time) > 0) {
      if (dpy->motion_buffer != 0) {
	hist = XGetMotionEvents (dpy, 
				 ev->xmotion.root,
				 data->time+1, CurrentTime,
				 &nhist);
      }
      if (!hist) {
	Window root, child;
	int x, y;
	int xr, yr;
	unsigned int mask;
	XQueryPointer(dpy, w, &root, &child, 
		      &xr, &yr,
		      &x, &y, 
		      &mask);
	nohist.x = xr;
	nohist.y = yr;
	nohist.time = ev->xmotion.time;
	hist = &nohist;
	nhist = 1;
      }
      hend = &hist[nhist];
      for (hptr = hist; hptr<hend; hptr++) {
	n = data->curr->n;
	if (n >= NPOLYLINE) break;

	hnxt = hptr + 1;
	if (hnxt < hend && 
	    hptr->time == hnxt->time)  continue;

	n = data->curr->n;
	if (n >= NPOLYLINE) break;

	XDrawString (dpy, ev->xmotion.root, cross, 
		 hptr->x, hptr->y, ".",1);

	data->curr->pts[n].x = hptr->x 
	                         - ev->xmotion.x_root
				 + ev->xmotion.x;
	data->curr->pts[n].y = hptr->y
	                         - ev->xmotion.y_root
				 + ev->xmotion.y;

	XDrawLines (dpy, w, gc, &(data->curr->pts[n-1]), 2,
		    CoordModeOrigin);

	data->curr->n = n+1;
      }
      data->time = hist[nhist-1].time;
      if (hist != &nohist) XFree (hist);
    }
    break;

  case ButtonRelease:
    XDrawString (dpy, ev->xbutton.root, cross, 
		 ev->xbutton.x_root, ev->xbutton.y_root, "R",1);

    if (!data->curr) break;
    n = data->curr->n;
    if (n < NPOLYLINE) {
      data->curr->pts[n].x = ev->xbutton.x;
      data->curr->pts[n].y = ev->xbutton.y;
      XDrawLines (dpy, w, gc, &(data->curr->pts[n-1]), 2,
		  CoordModeOrigin);
      data->curr->n = n + 1;

    }
    data->curr->next = data->line;
    data->line = data->curr;
    data->curr = 0;
    break;
  default:
    break;
  }
}


/* ARGSUSED */
void motion_keydown ( widget, data, ev )
  Widget widget;
  motion_data_t * data;
  XEvent * ev;
{
  exit (0);
}

